@@ -36,22 +36,37 @@
 #include "pem.h"
 #include "tls-private.h"
 #include "key.h"
+#include "asn1-private.h"
 
-void tls10_prf(const uint8_t *secret, size_t secret_len,
+void tls10_prf(const void *secret, size_t secret_len,
 		const char *label,
-		const uint8_t *seed, size_t seed_len,
+		const void *seed, size_t seed_len,
 		uint8_t *out, size_t out_len)
 {
 	uint8_t p_hash2[out_len];
 	uint8_t l_s1 = (secret_len + 1) / 2;
 	unsigned int i;
 
+	/*
+	 * RFC2246 section 5:
+	 * S1 and S2 are the two halves of the secret, and each is the same
+	 * length.  S1 is taken from the first half of the secret, S2 from the
+	 * second half.  Their length is created by rounding up the length of
+	 * the overall secret, divided by two; thus, if the original secret is
+	 * an odd number of bytes long, the last byte of S1 will be the same as
+	 * the first byte of S2.
+	 */
+
 	tls12_prf(L_CHECKSUM_MD5, 16,
 			secret, l_s1,
 			label, seed, seed_len,
 			out, out_len);
+
+	if (secret_len > 0)
+		secret += secret_len - l_s1;
+
 	tls12_prf(L_CHECKSUM_SHA1, 20,
-			secret + secret_len - l_s1, l_s1,
+			secret, l_s1,
 			label, seed, seed_len,
 			p_hash2, out_len);
 
@@ -60,9 +75,9 @@ void tls10_prf(const uint8_t *secret, size_t secret_len,
 }
 
 void tls12_prf(enum l_checksum_type type, size_t hash_len,
-		const uint8_t *secret, size_t secret_len,
+		const void *secret, size_t secret_len,
 		const char *label,
-		const uint8_t *seed, size_t seed_len,
+		const void *seed, size_t seed_len,
 		uint8_t *out, size_t out_len)
 {
 	struct l_checksum *hmac = l_checksum_new_hmac(type, secret, secret_len);
@@ -99,9 +114,9 @@ void tls12_prf(enum l_checksum_type type, size_t hash_len,
 
 void tls_prf_get_bytes(struct l_tls *tls,
 				enum l_checksum_type type, size_t hash_len,
-				const uint8_t *secret, size_t secret_len,
+				const void *secret, size_t secret_len,
 				const char *label,
-				const uint8_t *seed, size_t seed_len,
+				const void *seed, size_t seed_len,
 				uint8_t *buf, size_t len)
 {
 	if (tls->negotiated_version >= TLS_V12)
@@ -111,6 +126,32 @@ void tls_prf_get_bytes(struct l_tls *tls,
 		tls10_prf(secret, secret_len, label, seed, seed_len, buf, len);
 }
 
+LIB_EXPORT bool l_tls_prf_get_bytes(struct l_tls *tls,
+				enum l_checksum_type type,
+				size_t hash_len, bool use_master_secret,
+				const char *label, uint8_t *buf, size_t len)
+{
+	uint8_t seed[64];
+
+	if (unlikely(!tls))
+		return false;
+
+	memcpy(seed +  0, tls->pending.client_random, 32);
+	memcpy(seed + 32, tls->pending.server_random, 32);
+
+	if (use_master_secret)
+		tls_prf_get_bytes(tls, type, hash_len,
+						tls->pending.master_secret, 48,
+						label, seed, 64, buf, len);
+	else
+		tls_prf_get_bytes(tls, type, hash_len, "", 0,
+						label, seed, 64, buf, len);
+
+	memset(seed, 0, 64);
+
+	return true;
+}
+
 static void tls_write_random(uint8_t *buf)
 {
 	l_put_be32(time(NULL), buf);
@@ -1417,37 +1458,23 @@ decode_error:
 static void tls_handle_certificate(struct l_tls *tls,
 					const uint8_t *buf, size_t len)
 {
-	int total, cert_len;
-	struct tls_cert *certchain = NULL, **tail = &certchain;
+	size_t total;
+	struct tls_cert *certchain = NULL;
 	struct tls_cert *ca_cert = NULL;
 	bool dummy;
 
-	/* Length checks */
+	if (len < 3)
+		goto decode_error;
 
+	/* Length checks */
 	total = *buf++ << 16;
 	total |= *buf++ << 8;
 	total |= *buf++ << 0;
-	if ((size_t) total + 3 != len)
+	if (total + 3 != len)
 		goto decode_error;
 
-	while (total) {
-		cert_len = *buf++ << 16;
-		cert_len |= *buf++ << 8;
-		cert_len |= *buf++ << 0;
-
-		if (cert_len + 3 > total)
-			goto decode_error;
-
-		*tail = l_malloc(sizeof(struct tls_cert) + cert_len);
-		(*tail)->size = cert_len;
-		(*tail)->issuer = NULL;
-		memcpy((*tail)->asn1, buf, cert_len);
-
-		tail = &(*tail)->issuer;
-
-		buf += cert_len;
-		total -= cert_len + 3;
-	}
+	if (tls_cert_from_certificate_list(buf, total, &certchain) < 0)
+		goto decode_error;
 
 	/*
 	 * "Note that a client MAY send no certificates if it does not have any
@@ -2217,19 +2244,20 @@ LIB_EXPORT bool l_tls_set_auth_data(struct l_tls *tls, const char *cert_path,
 	}
 
 	if (priv_key_path) {
-		uint8_t *priv_key = NULL;
+		uint8_t *priv_key;
 		bool is_public = true;
 
 		priv_key = l_pem_load_private_key(priv_key_path,
 							priv_key_passphrase,
+							NULL,
 							&tls->priv_key_size);
+		if (!priv_key)
+			return false;
 
 		tls->priv_key = l_key_new(L_KEY_RSA, priv_key,
 						tls->priv_key_size);
-		if (priv_key) {
-			memset(priv_key, 0, tls->priv_key_size);
-			l_free(priv_key);
-		}
+		memset(priv_key, 0, tls->priv_key_size);
+		l_free(priv_key);
 
 		if (!l_key_get_info(tls->priv_key, L_KEY_RSA_PKCS1_V1_5,
 					L_CHECKSUM_NONE, &tls->priv_key_size,
@@ -2309,19 +2337,6 @@ LIB_EXPORT const char *l_tls_alert_to_str(enum l_tls_alert_desc desc)
 
 /* X509 Certificates and Certificate Chains */
 
-#define ASN1_ID(class, pc, tag)	(((class) << 6) | ((pc) << 5) | (tag))
-
-#define ASN1_CLASS_UNIVERSAL	0
-
-#define ASN1_ID_SEQUENCE	ASN1_ID(ASN1_CLASS_UNIVERSAL, 1, 0x10)
-#define ASN1_ID_SET		ASN1_ID(ASN1_CLASS_UNIVERSAL, 1, 0x11)
-#define ASN1_ID_INTEGER		ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x02)
-#define ASN1_ID_BIT_STRING	ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x03)
-#define ASN1_ID_OCTET_STRING	ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x04)
-#define ASN1_ID_OID		ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x06)
-#define ASN1_ID_UTF8STRING	ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x0c)
-#define ASN1_ID_PRINTABLESTRING	ASN1_ID(ASN1_CLASS_UNIVERSAL, 0, 0x13)
-
 #define X509_CERTIFICATE_POS			0
 #define   X509_TBSCERTIFICATE_POS		  0
 #define     X509_TBSCERT_VERSION_POS		    0
@@ -2341,83 +2356,6 @@ LIB_EXPORT const char *l_tls_alert_to_str(enum l_tls_alert_desc desc)
 #define   X509_SIGNATURE_ALGORITHM_POS		  1
 #define   X509_SIGNATURE_VALUE_POS		  2
 
-static inline int parse_asn1_definite_length(const uint8_t **buf,
-						size_t *len)
-{
-	int n;
-	size_t result = 0;
-
-	(*len)--;
-
-	if (!(**buf & 0x80))
-		return *(*buf)++;
-
-	n = *(*buf)++ & 0x7f;
-	if ((size_t) n > *len)
-		return -1;
-
-	*len -= n;
-	while (n--)
-		result = (result << 8) | *(*buf)++;
-
-	return result;
-}
-
-/* Return index'th element in a DER SEQUENCE */
-static uint8_t *der_find_elem(uint8_t *buf, size_t len_in, int index,
-			uint8_t *tag, size_t *len_out)
-{
-	int tlv_len;
-
-	while (1) {
-		if (len_in < 2)
-			return NULL;
-
-		*tag = *buf++;
-		len_in--;
-
-		tlv_len = parse_asn1_definite_length((void *) &buf, &len_in);
-		if (tlv_len < 0 || (size_t) tlv_len > len_in)
-			return NULL;
-
-		if (index-- == 0) {
-			*len_out = tlv_len;
-			return buf;
-		}
-
-		buf += tlv_len;
-		len_in -= tlv_len;
-	}
-}
-
-/* Return an element in a DER SEQUENCE structure by path */
-static inline uint8_t *der_find_elem_by_path(uint8_t *buf, size_t len_in,
-						uint8_t tag, size_t *len_out,
-						...)
-{
-	uint8_t elem_tag;
-	int pos;
-	va_list vl;
-
-	va_start(vl, len_out);
-
-	pos = va_arg(vl, int);
-
-	while (pos != -1) {
-		buf = der_find_elem(buf, len_in, pos, &elem_tag, &len_in);
-
-		pos = va_arg(vl, int);
-
-		if (!buf || elem_tag != (pos == -1 ? tag : ASN1_ID_SEQUENCE))
-			return NULL;
-	}
-
-	va_end(vl);
-
-	*len_out = len_in;
-	return buf;
-}
-
 struct tls_cert *tls_cert_load_file(const char *filename)
 {
 	uint8_t *der;
@@ -2443,17 +2381,52 @@ struct tls_cert *tls_cert_load_file(const char *filename)
 	return cert;
 }
 
+int tls_cert_from_certificate_list(const void *data, size_t len,
+					struct tls_cert **out_certchain)
+{
+	const uint8_t *buf = data;
+	struct tls_cert *certchain = NULL, **tail = &certchain;
+
+	while (len) {
+		size_t cert_len;
+
+		if (len < 3)
+			goto decode_error;
+
+		cert_len = *buf++ << 16;
+		cert_len |= *buf++ << 8;
+		cert_len |= *buf++ << 0;
+
+		if (cert_len + 3 > len)
+			goto decode_error;
+
+		*tail = l_malloc(sizeof(struct tls_cert) + cert_len);
+		(*tail)->size = cert_len;
+		(*tail)->issuer = NULL;
+		memcpy((*tail)->asn1, buf, cert_len);
+
+		tail = &(*tail)->issuer;
+
+		buf += cert_len;
+		len -= cert_len + 3;
+	}
+
+	if (out_certchain)
+		*out_certchain = certchain;
+
+	return 0;
+
+decode_error:
+	tls_cert_free_certchain(certchain);
+	return -EBADMSG;
+}
+
 bool tls_cert_find_certchain(struct tls_cert *cert,
 				const char *cacert_filename)
 {
 	return true;
 }
 
-struct asn1_oid {
-	uint8_t asn1_len;
-	uint8_t asn1[10];
-};
-
 static const struct pkcs1_encryption_oid {
 	enum tls_cert_key_type key_type;
 	struct asn1_oid oid;
@@ -2583,12 +2556,12 @@ void tls_cert_free_certchain(struct tls_cert *cert)
 
 enum tls_cert_key_type tls_cert_get_pubkey_type(struct tls_cert *cert)
 {
-	uint8_t *key_type;
+	const uint8_t *key_type;
 	size_t key_type_len;
 	int i;
 
-	key_type = der_find_elem_by_path(cert->asn1, cert->size, ASN1_ID_OID,
-						&key_type_len,
+	key_type = asn1_der_find_elem_by_path(cert->asn1, cert->size,
+						ASN1_ID_OID, &key_type_len,
 						X509_CERTIFICATE_POS,
 						X509_TBSCERTIFICATE_POS,
 						X509_TBSCERT_SUBJECT_KEY_POS,
@@ -2599,10 +2572,8 @@ enum tls_cert_key_type tls_cert_get_pubkey_type(struct tls_cert *cert)
 		return TLS_CERT_KEY_UNKNOWN;
 
 	for (i = 0; i < (int) L_ARRAY_SIZE(pkcs1_encryption_oids); i++)
-		if (key_type_len == pkcs1_encryption_oids[i].oid.asn1_len &&
-				!memcmp(key_type,
-					pkcs1_encryption_oids[i].oid.asn1,
-					key_type_len))
+		if (asn1_oid_eq(&pkcs1_encryption_oids[i].oid,
+					key_type_len, key_type))
 			break;
 
 	if (i == L_ARRAY_SIZE(pkcs1_encryption_oids))
